# -*- coding: utf-8 -*-
import os

from contextlib import contextmanager


@contextmanager
def open_here(filepath, *args, **kwargs):
    """open a file relative to this files location"""

    here = os.path.abspath(os.path.dirname(__file__))
    fp = open(os.path.join(here, filepath), *args, **kwargs)
    yield fp
    fp.close()


def up(cursor):
    """Add subcol metadata support to shredding and data upgrades"""

    cursor.execute("""
CREATE FUNCTION uuid5(namespace uuid, name text) RETURNS uuid
    LANGUAGE plpythonu
    AS $_$ import uuid; return uuid.uuid5(uuid.UUID(namespace), name) $_$;""")

    cursor.execute("""
CREATE OR REPLACE FUNCTION public.is_baked(col_uuid uuid, col_ver text)
 RETURNS boolean
 LANGUAGE sql
AS $function$
SELECT bool_or(is_collated)
    FROM modules JOIN trees
        ON module_ident = documentid
    WHERE uuid = col_uuid AND
          module_version(major_version, minor_version) = col_ver
$function$;""")

    with open_here('../archive-sql/schema/subcol_uuids_func.sql', 'rb') as f:
        cursor.execute(f.read())

    with open_here('../archive-sql/schema/shred_collxml.sql', 'rb') as f:
        cursor.execute(f.read())


def down(cursor):
    # TODO rollback code

    cursor.execute('drop function uuid5(uuid, text)')
    cursor.execute('drop function is_baked(uuid, text)')
    cursor.execute('drop function subcol_uuids(text, text)')

    with open_here('shred_collxml_20170201213850_pre.sql', 'rb') as f:
        cursor.execute(f.read())



# -*- coding: utf-8 -*-
"""Add a trigger on modules to set all new inserted books to have state
"post-publication" in case there is a ruleset for collation.
"""


def up(cursor):
    cursor.execute("""\
CREATE OR REPLACE FUNCTION update_default_modules_stateid ()
RETURNS TRIGGER
LANGUAGE PLPGSQL
AS $$
BEGIN
  IF NEW.portal_type = 'Collection' THEN
    NEW.stateid = 5;
  END IF;
  RETURN NEW;
END
$$;

CREATE TRIGGER update_default_modules_stateid
  BEFORE INSERT ON modules FOR EACH ROW
  EXECUTE PROCEDURE update_default_modules_stateid();""")


def down(cursor):
    cursor.execute(
        'DROP TRIGGER IF EXISTS update_default_modules_stateid ON modules')
    cursor.execute('DROP FUNCTION IF EXISTS update_default_modules_stateid()')



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
        ALTER TABLE pending_resources
            ADD COLUMN filename TEXT""")


def down(cursor):
    cursor.execute("""\
        ALTER TABLE pending_resources
            DROP COLUMN filename""")



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
ALTER TABLE modules
  ALTER COLUMN stateid SET DEFAULT 1""")


def down(cursor):
    cursor.execute("""\
ALTER TABLE modules
  ALTER COLUMN stateid DROP DEFAULT""")



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
CREATE OR REPLACE FUNCTION post_publication() RETURNS trigger AS $$
BEGIN
  NOTIFY post_publication;
  RETURN NEW;
END;
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER post_publication_trigger
  AFTER INSERT OR UPDATE ON modules FOR EACH ROW
  WHEN (NEW.stateid = 5)
  EXECUTE PROCEDURE post_publication();""")


def down(cursor):
    cursor.execute("""\
DROP TRIGGER post_publication_trigger ON modules;
DROP FUNCTION post_publication();""")



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
CREATE TYPE post_publication_states AS ENUM (
  'Done/Success',
  'Failed/Error',
  'Processing'
);
CREATE TABLE post_publications (
  "module_ident" INTEGER NOT NULL,
  "state" post_publication_states NOT NULL,
  "state_message" TEXT,
  "timestamp" TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
  FOREIGN KEY ("module_ident") REFERENCES modules ("module_ident")
);""")


def down(cursor):
    cursor.execute('DROP TYPE IF EXISTS collation_states')
    cursor.execute('DROP TABLE IF EXISTS collation_states')



# -*- coding: utf-8 -*-
"""\
- Migrate all `module_files` records to a single `files` entry.
- Fix the `files` table to only allow one copy of a file.

"""


def up(cursor):
    # Migrate all `module_files` records to a single `files` entry.
    cursor.execute("""\
WITH grouped_files AS (
  SELECT f.sha1, array_agg(f.fileid ORDER BY f.fileid ASC) AS fileids
  FROM files AS f
  WHERE (SELECT DISTINCT count(*) FROM files AS f2 WHERE f.sha1 = f2.sha1) > 1
  GROUP BY f.sha1
),
for_removal AS (
  SELECT unnest(gf.fileids[2:array_length(gf.fileids,1)]) AS fileid
  FROM grouped_files as gf
),
updated_files as (
  UPDATE module_files AS mf
  SET fileid = gf.fileids[1]
  FROM grouped_files AS gf
  WHERE mf.fileid = any(gf.fileids)
  RETURNING mf.fileid
)
DELETE FROM files WHERE fileid IN (SELECT fileid FROM for_removal)
""")
    # Fix the `files` table to only allow one copy of a file.
    cursor.execute("ALTER TABLE files ADD UNIQUE (sha1)")


def down(cursor):
    pass



# -*- coding: utf-8 -*-
import os
import sys

import psycopg2
import pytest

from .. import testing


def _translate_parts_to_args(parts):
    """Translates connection string parts to arguments"""
    return ['-h', parts['host'],
            '-p', parts['port'],
            '-d', parts['dbname'],
            '-U', parts['user'],
            ]


@pytest.mark.usefixtures('db_wipe')
def test_init(connection_string_parts):
    from cnxdb.cli.main import main
    args = ['init'] + _translate_parts_to_args(connection_string_parts)
    return_code = main(args)

    assert return_code == 0

    def table_name_filter(table_name):
        return (not table_name.startswith('pg_') and
                not table_name.startswith('_pg_'))

    with psycopg2.connect(**connection_string_parts) as conn:
        with conn.cursor() as cursor:
            cursor.execute("SELECT table_name "
                           "FROM information_schema.tables "
                           "ORDER BY table_name")
            tables = [table_name for (table_name,) in cursor.fetchall()
                      if table_name_filter(table_name)]

    assert 'modules' in tables
    assert 'pending_documents' in tables


@pytest.mark.usefixtures('db_wipe')
def test_init_called_twice(capsys, connection_string_parts):
    from cnxdb.cli.main import main
    args = ['init'] + _translate_parts_to_args(connection_string_parts)

    return_code = main(args)
    assert return_code == 0

    return_code = main(args)
    assert return_code == 3
    out, err = capsys.readouterr()
    assert 'already initialized' in err


@pytest.mark.skipif(not testing.is_db_local(),
                    reason="not testing against a local database")
@pytest.mark.usefixtures('db_wipe')
def test_init_local(connection_string_parts):
    from cnxdb.cli.main import main
    args = ['init'] + _translate_parts_to_args(connection_string_parts)[4:]

    return_code = main(args)
    assert return_code == 0


@pytest.mark.usefixtures('db_wipe')
def test_init_without_dbname(connection_string_parts):
    from cnxdb.cli.main import main
    args = ['init']
    args.extend(_translate_parts_to_args(connection_string_parts)[:4])
    args.extend(_translate_parts_to_args(connection_string_parts)[6:])

    with pytest.raises(SystemExit) as exc_info:
        return_code = main(args)
    assert exc_info.value.code == 2


@pytest.mark.usefixtures('db_wipe')
def test_init_without_user(connection_string_parts):
    from cnxdb.cli.main import main
    args = ['init'] + _translate_parts_to_args(connection_string_parts)[:6]

    with pytest.raises(SystemExit) as exc_info:
        return_code = main(args)
    assert exc_info.value.code == 2


def assert_venv_is_active(connection_string_parts):
    """Asserts the venv is active and working"""
    with psycopg2.connect(**connection_string_parts) as conn:
        with conn.cursor() as cursor:
            cursor.execute("CREATE OR REPLACE FUNCTION pyprefix() "
                           "RETURNS text LANGUAGE "
                           "plpythonu AS $$import sys;return sys.prefix$$")
            cursor.execute("SELECT pyprefix()")
            db_pyprefix = cursor.fetchone()[0]

    assert os.path.samefile(db_pyprefix, sys.prefix)


@pytest.mark.skipif(not testing.is_venv(), reason="not within a venv")
@pytest.mark.usefixtures('db_init_and_wipe')
def test_venv(connection_string_parts):
    # Remove the venv schema before trying to initialize it.
    with psycopg2.connect(**connection_string_parts) as conn:
        with conn.cursor() as cursor:
            cursor.execute("DROP SCHEMA venv CASCADE")

    from cnxdb.cli.main import main
    args = ['venv'] + _translate_parts_to_args(connection_string_parts)

    return_code = main(args)
    assert return_code == 0

    assert_venv_is_active(connection_string_parts)


@pytest.mark.skipif(not testing.is_venv(), reason="not within a venv")
@pytest.mark.usefixtures('db_init_and_wipe')
def test_venv_called_twice(connection_string_parts):
    # Note, the initialization already setup the venv,
    # so this really calles 3 times.
    from cnxdb.cli.main import main
    args = ['venv'] + _translate_parts_to_args(connection_string_parts)

    return_code = main(args)
    assert return_code == 0

    return_code = main(args)
    assert return_code == 0

    assert_venv_is_active(connection_string_parts)



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
        CREATE TABLE pending_resource_associations (
          "document_id" INTEGER NOT NULL,
          "resource_id" INTEGER NOT NULL,
          PRIMARY KEY ("document_id", "resource_id"),
          FOREIGN KEY ("document_id") REFERENCES pending_documents ("id"),
          FOREIGN KEY ("resource_id") REFERENCES pending_resources ("id")
        )""")


def down(cursor):
    cursor.execute('DROP TABLE pending_resource_associations')



# -*- coding: utf-8 -*-
"""cnx-db database control"""
from __future__ import print_function
import argparse

from .discovery import discover_subcommands


def create_main_parser():
    parser = argparse.ArgumentParser(__doc__)
    discover_subcommands(parser)
    return parser


def main(argv=None):
    parser = create_main_parser()
    args = parser.parse_args(argv)

    return args.cmd(args)


__all__ = ('main',)



# -*- coding: utf-8 -*-


class DBSchemaInitialized(Exception):
    """Raised when a database schema has been initialized and an another
    attempt is being made to initialize it.

    """


__all__ = ('DBSchemaInitialized',)



#!/usr/bin/env python
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""commandline tool for parsing collxml into a DB tree."""


from xml import sax
import sys
import psycopg2

# While the collxml files we process potentially contain many of these
# namespaces, I take advantage of the fact that almost none of the
# localnames (tags names) acutally overlap. The one case that does (title)
# actually works in our favor, since we want to treat it the same anyway.

ns = {"cnx": "http://cnx.rice.edu/cnxml",
      "cnxorg": "http://cnx.rice.edu/system-info",
      "md": "http://cnx.rice.edu/mdml",
      "col": "http://cnx.rice.edu/collxml",
      "cnxml": "http://cnx.rice.edu/cnxml",
      "m": "http://www.w3.org/1998/Math/MathML",
      "q": "http://cnx.rice.edu/qml/1.0",
      "xhtml": "http://www.w3.org/1999/xhtml",
      "bib": "http://bibtexml.sf.net/",
      "cc": "http://web.resource.org/cc/",
      "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#"}

NODE_INS = "INSERT INTO trees (parent_id,documentid,childorder) "\
    "SELECT %s, module_ident, %s from modules WHERE "\
    "moduleid = %s AND version = %s RETURNING nodeid"
NODE_NODOC_INS = "INSERT INTO trees (parent_id,childorder) VALUES (%s, %s) "\
    "RETURNING nodeid"
NODE_TITLE_UPD = "UPDATE trees SET title = %s FROM modules WHERE nodeid = %s "\
    "AND (documentid IS NULL "\
    "OR (documentid = module_ident AND name != %s))"

con = psycopg2.connect('dbname=repository')
cur = con.cursor()


def _do_insert(pid, cid, oid=0, ver=0):
    if oid:
        cur.execute(NODE_INS, (pid, cid, oid, ver))
        if cur.rowcount == 0:  # no documentid found
            cur.execute(NODE_NODOC_INS, (pid, cid))
    else:
        cur.execute(NODE_NODOC_INS, (pid, cid))
    res = cur.fetchall()
    if res:
        nodeid = res[0][0]
    else:
        nodeid = None
    return nodeid


def _do_update(title, nid):
    cur.execute(NODE_TITLE_UPD, (title, nid, title))


class ModuleHandler(sax.ContentHandler):
    """Handler for module link."""

    def __init__(self):
        """Create module handler with default values."""
        self.parents = [None]
        self.childorder = 0
        self.map = {}
        self.tag = u''
        self.contentid = u''
        self.version = u''
        self.title = u''
        self.nodeid = 0
        self.derivedfrom = [None]

    def startElementNS(self, (uri, localname), qname, attrs):
        """Handle element."""
        self.map[localname] = u''
        self.tag = localname

        if localname == 'module':
            self.childorder[-1] += 1
            nodeid = _do_insert(self.parents[-1], self.childorder[-1],
                                attrs[(None, "document")],
                                attrs[(ns["cnxorg"],
                                      "version-at-this-collection-version")])
            if nodeid:
                self.nodeid = nodeid

        elif localname == 'subcollection':
            # TODO insert a metadata record into modules table for subcol.
            self.childorder[-1] += 1
            nodeid = _do_insert(self.parents[-1], self.childorder[-1])
            if nodeid:
                self.nodeid = nodeid
                self.parents.append(self.nodeid)
            self.childorder.append(1)

        elif localname == 'derived-from':
            self.derivedfrom.append(True)

    def characters(self, content):
        """Copy characters to tag."""
        self.map[self.tag] += content

    def endElementNS(self, (uris, localname), qname):
        """Assign local values."""
        if localname == 'content-id' and not self.derivedfrom[-1]:
            self.contentid = self.map[localname]
        elif localname == 'version' and not self.derivedfrom[-1]:
            self.version = self.map[localname]
        elif localname == 'title' and not self.derivedfrom[-1]:
            self.title = self.map[localname]
            if self.parents[-1]:  # current node is a subcollection or module
                _do_update(self.title.encode('utf-8'), self.nodeid)

        elif localname == 'derived-from':
            self.derivedfrom.pop()

        elif localname == 'metadata':
            # We know that at end of metadata, we've got the collection info
            self.childorder = [0]
            nodeid = _do_insert(None, self.childorder[-1],
                                self.contentid, self.version)
            if nodeid:
                self.nodeid = nodeid
                self.parents.append(self.nodeid)
            self.childorder.append(1)

        elif localname == 'content':
            # this occurs at the end of each container class:
            # either colletion or subcollection
            self.parents.pop()
            self.childorder.pop()


parser = sax.make_parser()
parser.setFeature(sax.handler.feature_namespaces, 1)
parser.setContentHandler(ModuleHandler())
parser.parse(open(sys.argv[1]))

con.commit()



# -*- coding: utf-8 -*-
"""cnx-db subcommands"""
from __future__ import print_function
import sys

from .common import assign_db_args
from .discovery import register_subcommand


def _compile_connection_string_parts(args_namespace):
    """Given an ``argparse.Namespace``, translate this to a connection string
    parts (dict).

    """
    return {
        'dbname': args_namespace.dbname,
        'user': args_namespace.user,
        # 'password': None,
        'host': args_namespace.host,
        'port': args_namespace.port,
    }


def _translate_parts_to_string(connection_string_parts):
    """Translate the connection string parts ot a string"""
    return ' '.join(['='.join([k, v])
                     for k, v in connection_string_parts.items()
                     if v is not None])


@register_subcommand('init', *assign_db_args)
def init_cmd(args_namespace):
    """initialize the database"""
    connection_string_parts = _compile_connection_string_parts(args_namespace)
    connection_string = _translate_parts_to_string(connection_string_parts)
    from ..init import init_db, DBSchemaInitialized
    try:
        init_db(connection_string, False)
    except DBSchemaInitialized:
        print("Database is already initialized", file=sys.stderr)
        return 3
    return 0


@register_subcommand('venv', *assign_db_args)
def venv_cmd(args_namespace):
    """(re)initialize the venv within the database"""
    connection_string_parts = _compile_connection_string_parts(args_namespace)
    connection_string = _translate_parts_to_string(connection_string_parts)
    from ..init import init_venv
    init_venv(connection_string)
    return 0



# -*- coding: utf-8 -*-
import functools
import os
import sys

import psycopg2


_DEFAULT_CONNECTION_SETTINGS = {
    'dbname': 'testing',
    'user': 'tester',
    'password': 'tester',
    'host': 'localhost',
    'port': '5432',
    }


def parse_connection_string_to_parts(connection_string):
    """Parses a connection string to parts (dict of key values)."""
    return dict(map(lambda x: x.split('='), connection_string.split()))


def get_connection_string_parts():
    """Retrieves the connection string as dictionary key values."""
    parts = {}
    [parts.setdefault(k, v) for k, v in _DEFAULT_CONNECTION_SETTINGS.items()]
    connection_string = os.environ.get('TESTING_CONNECTION_STRING', None)
    if connection_string is not None:
        parts.update(parse_connection_string_to_parts(connection_string))
    return parts


def get_connection_string():
    """Retrieves the connection string from configuration."""
    return ' '.join(
        ['{}={}'.format(key, value)
         for key, value in get_connection_string_parts().items()])


def db_connection_factory(connection_string=None):
    if connection_string is None:
        connection_string = get_connection_string()

    def db_connect():
        return psycopg2.connect(connection_string)

    return db_connect


def db_connect(method):
    """Decorator for methods that need to use the database

    Example:
    @db_connect
    def setUp(self, cursor):
        cursor.execute(some_sql)
        # some other code
    """
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        connect = db_connection_factory()
        with connect() as db_connection:
            with db_connection.cursor() as cursor:
                return method(self, cursor, *args, **kwargs)
    return wrapped


def is_venv():
    """Returns a boolean telling whether the application is running
    within a virtualenv (aka venv).

    """
    return hasattr(sys, 'real_prefix')


def is_db_local():
    """Returns a boolean telling whether the database is local or not."""
    return get_connection_string_parts()['host'] == 'localhost'


__all__ = (
    'db_connect',
    'db_connection_factory',
    'get_connection_string',
    'is_db_local',
    'is_venv',
    )



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
CREATE OR REPLACE FUNCTION strip_html(html_text TEXT)
  RETURNS text
AS $$
  import re
  return re.sub('<[^>]*?>', '', html_text, re.MULTILINE)
$$ LANGUAGE plpythonu IMMUTABLE;
    """)


def down(cursor):
    cursor.execute("DROP FUNCTION IF EXISTS strip_html(TEXT)")



# -*- coding: utf-8 -*-


def add_db_args_to_parser(parser):
    """Adds the database arguments. These arguments are modeled after ``psql``'s
    arguments.

    """
    parser.add_argument('-h', '--host',
                        default='localhost',
                        help="database host name")
    parser.add_argument('-p', '--port',
                        default='5432',
                        help="database port")
    parser.add_argument('-d', '--dbname',
                        required=True,
                        help="database name")
    parser.add_argument('-U', '--user',
                        required=True,
                        help="database user")
    # parser.add_argument('-W', '--password', help="database password")


# Variable used to combined the parser callback and parser keyword arguments
# for use with the register_subcommand.
assign_db_args = (add_db_args_to_parser, {'add_help': False},)


__all__ = ('add_db_args_to_parser', 'assign_db_args',)



# -*- coding: utf-8 -*-
import pytest

from .testing import get_connection_string, get_connection_string_parts


@pytest.fixture
def connection_string_parts():
    """Returns a connection string as parts (dict)"""
    return get_connection_string_parts()


@pytest.fixture
def connection_string():
    """Returns a connection string"""
    return get_connection_string()


@pytest.fixture
def db_wipe(connection_string, request):
    """Cleans up the database after a test run"""
    import psycopg2

    def finalize():
        with psycopg2.connect(connection_string) as conn:
            with conn.cursor() as cursor:
                cursor.execute("DROP SCHEMA public CASCADE; "
                               "CREATE SCHEMA public")
                cursor.execute("DROP SCHEMA IF EXISTS venv CASCADE")
    request.addfinalizer(finalize)


@pytest.fixture
def db_init(connection_string):
    """Initializes the database"""
    from cnxdb.init.main import init_db
    init_db(connection_string, True)


@pytest.fixture
def db_init_and_wipe(db_init, db_wipe):
    """Combination of the initialization and wiping procedures."""
    pass



# -*- coding: utf-8 -*-


def up(cursor):
    # Create collated fulltext tables, parallel to module tables, with an extra 'context' param.

    # Create tables, index functions, and indexes
    cursor.execute("""
CREATE TABLE collated_fti_lexemes (
    item integer,
    context integer,
    lexeme text,
    positions integer[]);

CREATE TABLE collated_fti (
    item integer NOT NULL,
    context integer NOT NULL,
    module_idx pg_catalog.tsvector,
    fulltext text);

ALTER TABLE collated_fti ADD CONSTRAINT collated_fti_pkey PRIMARY KEY (item, context);
ALTER TABLE collated_fti ADD CONSTRAINT collated_fti_context_fkey FOREIGN KEY (context) REFERENCES modules (module_ident) ON DELETE CASCADE;
ALTER TABLE collated_fti ADD CONSTRAINT collated_fti_item_fkey FOREIGN KEY (item) REFERENCES modules (module_ident) ON DELETE CASCADE;

ALTER TABLE collated_fti_lexemes ADD CONSTRAINT collated_fti_lexemes_item_fkey FOREIGN KEY (item) REFERENCES modules (module_ident) ON DELETE CASCADE;
ALTER TABLE collated_fti_lexemes ADD CONSTRAINT collated_fti_lexemes_context_fkey FOREIGN KEY (context) REFERENCES modules (module_ident) ON DELETE CASCADE;

CREATE FUNCTION module_version(major integer, minor integer) RETURNS text
    LANGUAGE sql IMMUTABLE
    AS $_$
  SELECT concat_ws('.', major, minor) ;
$_$;

CREATE INDEX collated_fti_idx ON collated_fti USING gist (module_idx);

CREATE INDEX trees_doc_idx ON trees (documentid);

CREATE INDEX modules_uuid_version_idx ON modules (uuid, module_version(major_version, minor_version));

CREATE INDEX latest_modules_uuid_version_idx ON latest_modules (uuid, module_version(major_version, minor_version));

CREATE FUNCTION count_collated_lexemes(myident integer, bookident integer, mysearch text) RETURNS bigint
    LANGUAGE sql STABLE
    AS $_$
     select sum(array_length(positions,1))
            from collated_fti_lexemes,
                 regexp_split_to_table(strip(to_tsvector(mysearch))::text,' ') s
            where item = myident and context = bookident and lexeme = substr(s,2,length(s)-2)
$_$;
""")
    # Update json tree functions to use new module_version function (and perhaps the index)
    cursor.execute("""
CREATE OR REPLACE FUNCTION tree_to_json(uuid TEXT, version TEXT, as_collated BOOLEAN DEFAULT TRUE) RETURNS TEXT as $$
select string_agg(toc,'
'
) from (
WITH RECURSIVE t(node, title, path,value, depth, corder, is_collated) AS (
    SELECT nodeid, title, ARRAY[nodeid], documentid, 1, ARRAY[childorder],
           is_collated
    FROM trees tr, modules m
    WHERE m.uuid::text = $1 AND
          module_version( m.major_version, m.minor_version) = $2 AND
      tr.documentid = m.module_ident AND
      tr.parent_id IS NULL AND
      tr.is_collated = $3
UNION ALL
    SELECT c1.nodeid, c1.title, t.path || ARRAY[c1.nodeid], c1.documentid, t.depth+1, t.corder || ARRAY[c1.childorder], c1.is_collated /* Recursion */
    FROM trees c1 JOIN t ON (c1.parent_id = t.node)
    WHERE not nodeid = any (t.path) AND t.is_collated = c1.is_collated
)
SELECT
    REPEAT('    ', depth - 1) ||
    '{"id":"' || COALESCE(m.uuid::text,'subcol') || concat_ws('.','@'||m.major_version, m.minor_version) ||'",' ||
    '"shortId":"' || COALESCE(short_id(m.uuid),'subcol') || concat_ws('.','@'||m.major_version, m.minor_version) ||'",' ||
      '"title":'||to_json(COALESCE(title,name))||
      CASE WHEN (depth < lead(depth,1,0) over(w)) THEN ', "contents":['
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) = 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) over(w) - 1)
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) = 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) over(w) - 1)
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) != 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) over(w))||','
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) != 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) over(w))||','
           WHEN m.uuid IS NULL THEN ', "contents":[]},'
           ELSE '},' END
      AS "toc"
FROM t left join  modules m on t.value = m.module_ident
    WINDOW w as (ORDER BY corder) order by corder ) tree ;
$$ LANGUAGE SQL;

CREATE OR REPLACE FUNCTION tree_to_json_for_legacy(TEXT, TEXT) RETURNS TEXT AS $$
SELECT string_agg(toc,'
'
) FROM (
WITH RECURSIVE t(node, title, path,value, depth, corder) AS (
    SELECT nodeid, title, ARRAY[nodeid], documentid, 1, ARRAY[childorder]
    FROM trees tr, modules m
    WHERE m.uuid::text = $1 AND
          module_version( m.major_version, m.minor_version) = $2 AND
      tr.documentid = m.module_ident AND
      tr.parent_id IS NULL AND
      tr.is_collated = FALSE
UNION ALL
    SELECT c1.nodeid, c1.title, t.path || ARRAY[c1.nodeid], c1.documentid, t.depth+1, t.corder || ARRAY[c1.childorder] /* Recursion */
    FROM trees c1 JOIN t ON (c1.parent_id = t.node)
    WHERE NOT nodeid = ANY (t.path) AND c1.is_collated = FALSE
)
SELECT
    REPEAT('    ', depth - 1) || '{"id":"' || COALESCE(m.moduleid,'subcol') ||  '",' ||
      '"version":' || COALESCE('"'||m.version||'"', 'null') || ',' ||
      '"title":'||to_json(COALESCE(title,name))||
      CASE WHEN (depth < lead(depth,1,0) OVER(w)) THEN ', "contents":['
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) = 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w) - 1)
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) = 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w) - 1)
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) != 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w))||','
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) != 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w))||','
           WHEN m.uuid IS NULL THEN ', "contents":[]},'
           ELSE '},' END
      AS "toc"
FROM t LEFT JOIN modules m ON t.value = m.module_ident
    WINDOW w AS (ORDER BY corder) ORDER BY corder ) tree ;
$$ LANGUAGE SQL;
""")
    # Trigger functions
    cursor.execute("""
CREATE FUNCTION index_collated_fulltext_trigger() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
  DECLARE
    has_existing_record integer;
    _baretext text;
    _idx_vectors tsvector;
  BEGIN
    has_existing_record := (SELECT item, context FROM collated_fti WHERE item = NEW.item and context = NEW.context);
    _baretext := (SELECT xml_to_baretext(convert_from(f.file, 'UTF8')::xml)::text FROM files AS f WHERE f.fileid = NEW.fileid);
    _idx_vectors := to_tsvector(_baretext);

    IF has_existing_record IS NULL THEN
      INSERT INTO collated_fti (item, context, fulltext, module_idx)
        VALUES ( NEW.item, NEW.context,_baretext, _idx_vectors );
    ELSE
      UPDATE collated_fti SET (fulltext, module_idx) = ( _baretext, _idx_vectors )
        WHERE item = NEW.item and context = NEW.context;
    END IF;
    RETURN NEW;
  END;
$_$;

CREATE FUNCTION index_collated_fulltext_lexeme_update_trigger() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
  BEGIN

    DELETE from collated_fti_lexemes where item = NEW.item;

    INSERT into collated_fti_lexemes (item, context, lexeme, positions)
       (with lex as (SELECT regexp_split_to_table(NEW.module_idx::text, E' \\'') as t )
       SELECT NEW.item, NEW.context,
              substring(t,1,strpos(t,E'\\':')-1),
              ('{'||substring(t,strpos(t,E'\\':')+2)||'}')::int[] from lex) ;

  RETURN NEW;
  END;
$_$;
""")
    # and the triggers
    cursor.execute("""
CREATE TRIGGER index_collated_fulltext
    AFTER INSERT OR UPDATE ON collated_file_associations
    FOR EACH ROW
    EXECUTE PROCEDURE index_collated_fulltext_trigger();

CREATE TRIGGER index_collated_fulltext_lexeme
    BEFORE INSERT OR UPDATE ON collated_fti
    FOR EACH ROW
    EXECUTE PROCEDURE index_collated_fulltext_lexeme_update_trigger();
""")

    # Clean up redundant lexeme triggers
    cursor.execute("""
DROP TRIGGER index_fulltext_lexeme ON modulefti;

CREATE TRIGGER index_fulltext_lexeme
    BEFORE INSERT OR UPDATE ON modulefti
    FOR EACH ROW
    EXECUTE PROCEDURE index_fulltext_lexeme_update_trigger();

DROP TRIGGER index_fulltext_lexeme_update ON modulefti;

DROP FUNCTION index_fulltext_lexeme_trigger();
""")

    # Fill tables
    cursor.execute("""
INSERT INTO collated_fti (item, context, fulltext, module_idx)
    SELECT item, context,
    xml_to_baretext(convert_from(f.file, 'UTF8')::xml)::text,
    to_tsvector(xml_to_baretext(convert_from(f.file, 'UTF8')::xml)::text)
    FROM files AS f, collated_file_associations AS cfa
    WHERE f.fileid = cfa.fileid
""")


def down(cursor):

    cursor.execute("""
CREATE FUNCTION index_fulltext_lexeme_trigger() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
  BEGIN

    INSERT into modulefti_lexemes (module_ident, lexeme, positions)
        (with lex as (SELECT regexp_split_to_table(NEW.module_idx::text, E' \\'') as t )
        SELECT NEW.module_ident,
               substring(t,1,strpos(t,E'\\':')-1),
               ('{'||substring(t,strpos(t,E'\\':')+2)||'}')::int[] from lex) ;

  RETURN NEW;
  END;
$_$;

CREATE TRIGGER index_fulltext_lexeme_update
    BEFORE UPDATE ON modulefti
    FOR EACH ROW
    EXECUTE PROCEDURE index_fulltext_lexeme_update_trigger();

DROP TRIGGER index_fulltext_lexeme ON modulefti;

CREATE TRIGGER index_fulltext_lexeme
    BEFORE INSERT ON modulefti
    FOR EACH ROW
    EXECUTE PROCEDURE index_fulltext_lexeme_trigger();
    """)

    cursor.execute("""
CREATE OR REPLACE FUNCTION tree_to_json(uuid TEXT, version TEXT, as_collated BOOLEAN DEFAULT TRUE) RETURNS TEXT as $$
select string_agg(toc,'
'
) from (
WITH RECURSIVE t(node, title, path,value, depth, corder, is_collated) AS (
    SELECT nodeid, title, ARRAY[nodeid], documentid, 1, ARRAY[childorder],
           is_collated
    FROM trees tr, modules m
    WHERE m.uuid::text = $1 AND
          concat_ws('.', m.major_version, m.minor_version) = $2 AND
      tr.documentid = m.module_ident AND
      tr.parent_id IS NULL AND
      tr.is_collated = $3
UNION ALL
    SELECT c1.nodeid, c1.title, t.path || ARRAY[c1.nodeid], c1.documentid, t.depth+1, t.corder || ARRAY[c1.childorder], c1.is_collated /* Recursion */
    FROM trees c1 JOIN t ON (c1.parent_id = t.node)
    WHERE not nodeid = any (t.path) AND t.is_collated = c1.is_collated
)
SELECT
    REPEAT('    ', depth - 1) ||
    '{"id":"' || COALESCE(m.uuid::text,'subcol') ||concat_ws('.', '@'||m.major_version, m.minor_version) ||'",' ||
    '"shortId":"' || COALESCE(short_id(m.uuid),'subcol') ||concat_ws('.', '@'||m.major_version, m.minor_version) ||'",' ||
      '"title":'||to_json(COALESCE(title,name))||
      CASE WHEN (depth < lead(depth,1,0) over(w)) THEN ', "contents":['
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) = 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) over(w) - 1)
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) = 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) over(w) - 1)
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) != 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) over(w))||','
           WHEN (depth > lead(depth,1,0) over(w) AND lead(depth,1,0) over(w) != 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) over(w))||','
           WHEN m.uuid IS NULL THEN ', "contents":[]},'
           ELSE '},' END
      AS "toc"
FROM t left join  modules m on t.value = m.module_ident
    WINDOW w as (ORDER BY corder) order by corder ) tree ;
$$ LANGUAGE SQL;

CREATE OR REPLACE FUNCTION tree_to_json_for_legacy(TEXT, TEXT) RETURNS TEXT AS $$
SELECT string_agg(toc,'
'
) FROM (
WITH RECURSIVE t(node, title, path,value, depth, corder) AS (
    SELECT nodeid, title, ARRAY[nodeid], documentid, 1, ARRAY[childorder]
    FROM trees tr, modules m
    WHERE m.uuid::text = $1 AND
          concat_ws('.',  m.major_version, m.minor_version) = $2 AND
      tr.documentid = m.module_ident AND
      tr.parent_id IS NULL AND
      tr.is_collated = FALSE
UNION ALL
    SELECT c1.nodeid, c1.title, t.path || ARRAY[c1.nodeid], c1.documentid, t.depth+1, t.corder || ARRAY[c1.childorder] /* Recursion */
    FROM trees c1 JOIN t ON (c1.parent_id = t.node)
    WHERE NOT nodeid = ANY (t.path) AND c1.is_collated = FALSE
)
SELECT
    REPEAT('    ', depth - 1) || '{"id":"' || COALESCE(m.moduleid,'subcol') ||  '",' ||
      '"version":' || COALESCE('"'||m.version||'"', 'null') || ',' ||
      '"title":'||to_json(COALESCE(title,name))||
      CASE WHEN (depth < lead(depth,1,0) OVER(w)) THEN ', "contents":['
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) = 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w) - 1)
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) = 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w) - 1)
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) != 0 AND m.uuid IS NULL) THEN ', "contents":[]}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w))||','
           WHEN (depth > lead(depth,1,0) OVER(w) AND lead(depth,1,0) OVER(w) != 0 ) THEN '}'||REPEAT(']}',depth - lead(depth,1,0) OVER(w))||','
           WHEN m.uuid IS NULL THEN ', "contents":[]},'
           ELSE '},' END
      AS "toc"
FROM t LEFT JOIN modules m ON t.value = m.module_ident
    WINDOW w AS (ORDER BY corder) ORDER BY corder ) tree ;
$$ LANGUAGE SQL;
""")

    cursor.execute("""
DROP TABLE collated_fti_lexemes;

DROP TABLE collated_fti;

DROP TRIGGER index_collated_fulltext ON collated_file_associations;

DROP FUNCTION index_collated_fulltext_trigger();

DROP FUNCTION index_collated_fulltext_lexeme_update_trigger();

DROP FUNCTION count_collated_lexemes(myident integer, bookident integer, mysearch text);

DROP INDEX trees_doc_idx;

DROP INDEX modules_uuid_version_idx;

DROP INDEX latest_modules_uuid_version_idx;

DROP FUNCTION module_version(major integer, minor integer);
""")



# -*- coding: utf-8 -*-
"""\
- Move the mimetype value from ``module_files`` to ``files``.
- Remove the ``mimetype`` column from the ``module_files`` table.

"""
from __future__ import print_function
import sys


def up(cursor):
    # Move the mimetype value from ``module_files`` to ``files``.
    cursor.execute("UPDATE files AS f SET media_type = mf.mimetype "
                   "FROM module_files AS mf "
                   "WHERE mf.fileid = f.fileid AND f.media_type IS NULL")

    # Warn about missing mimetype.
    cursor.execute("SELECT fileid, sha1 "
                   "FROM files AS f "
                   "WHERE f.fileid NOT IN (SELECT fileid FROM module_files)")
    rows = '\n'.join(['{}, {}'.format(fid, sha1)
                      for fid, sha1 in cursor.fetchall()])
    print("These files (fileid, sha1) do not have a corresponding "
          "module_files entry:\n{}\n".format(rows),
          file=sys.stderr)

    # Remove the ``mimetype`` column from the ``module_files`` table.
    cursor.execute("ALTER TABLE module_files DROP COLUMN mimetype")


def down(cursor):
    # Add a ``mimetype`` column to the ``module_files`` table.
    cursor.execute("ALTER TABLE module_files ADD COLUMN mimetype TEXT")

    # Move the mimetype value from ``files`` to ``module_files``.
    print("Rollback cannot accurately replace mimetype values that "
          "were in the ``modules_files`` table.",
          file=sys.stderr)
    cursor.execute("UPDATE module_files AS mf SET mimetype = f.media_type "
                   "FROM files AS f "
                   "WHERE f.fileid = mf.fileid")



# -*- coding: utf-8 -*-
import os

from contextlib import contextmanager


@contextmanager
def open_here(filepath, *args, **kwargs):
    """open a file relative to this files location"""

    here = os.path.abspath(os.path.dirname(__file__))
    fp = open(os.path.join(here, filepath), *args, **kwargs)
    yield fp
    fp.close()


def up(cursor):

    with open_here('../archive-sql/schema/functions.sql', 'rb') as f:
        cursor.execute(f.read())

    cursor.execute("""
CREATE INDEX modules_ident_hash on modules(ident_hash(uuid, major_version, minor_version));
CREATE INDEX modules_short_ident_hash on modules(short_ident_hash(uuid, major_version, minor_version));""")


def down(cursor):
    # TODO rollback code

    cursor.execute('drop function ident_hash(uuid, int, int) CASCADE')
    cursor.execute('drop function short_ident_hash(uuid, int, int) CASCADE')



# -*- coding: utf-8 -*-
"""\
- Add a ``media_type`` column to the ``files`` table.
- Move the mimetype value from ``module_files`` to ``files``.

"""
from __future__ import print_function
import sys


def up(cursor):
    # Add a ``media_type`` column to the ``files`` table.
    cursor.execute("ALTER TABLE files ADD COLUMN media_type TEXT")

    # Move the mimetype value from ``module_files`` to ``files``.
    cursor.execute("UPDATE files AS f SET media_type = mf.mimetype "
                   "FROM module_files AS mf "
                   "WHERE mf.fileid = f.fileid")

    # Warn about missing mimetype.
    cursor.execute("SELECT fileid, sha1 "
                   "FROM files AS f "
                   "WHERE f.fileid NOT IN (SELECT fileid FROM module_files)")
    rows = '\n'.join(['{}, {}'.format(fid, sha1)
                      for fid, sha1 in cursor.fetchall()])
    print("These files (fileid, sha1) do not have a corresponding "
          "module_files entry:\n{}\n".format(rows),
          file=sys.stderr)


def down(cursor):
    # Remove the ``mimetype`` column from the ``files`` table.
    cursor.execute("ALTER TABLE files DROP COLUMN media_type")



# -*- coding: utf-8 -*-
"""\
Change file hash update triggers to only trigger on file update
"""


def up(cursor):
    # Drop the triggers from the files table
    cursor.execute("DROP TRIGGER update_file_md5 ON files")
    cursor.execute("DROP TRIGGER update_files_sha1 ON files")

    # Create triggers that run only on file update
    cursor.execute("CREATE TRIGGER update_file_md5 "
                   "BEFORE INSERT OR UPDATE OF file ON files "
                   "FOR EACH ROW "
                   "EXECUTE PROCEDURE update_md5()")

    cursor.execute("CREATE TRIGGER update_files_sha1 "
                   "BEFORE INSERT OR UPDATE OF file ON files "
                   "FOR EACH ROW "
                   "EXECUTE PROCEDURE update_sha1()")


def down(cursor):
    # Drop the triggers from the files table
    cursor.execute("DROP TRIGGER update_file_md5 ON files")
    cursor.execute("DROP TRIGGER update_files_sha1 ON files")

    # Create triggers that run every time the files table is updated
    cursor.execute("CREATE TRIGGER update_file_md5 "
                   "BEFORE INSERT OR UPDATE ON files "
                   "FOR EACH ROW "
                   "EXECUTE PROCEDURE update_md5()")

    cursor.execute("CREATE TRIGGER update_files_sha1 "
                   "BEFORE INSERT OR UPDATE ON files "
                   "FOR EACH ROW "
                   "EXECUTE PROCEDURE update_sha1()")



# -*- coding: utf-8 -*-
from cnxdb.cli.discovery import register_subcommand


def assign_args(parser):
    parser.add_argument('-h', '--humph', action='store_true')
    parser.add_argument('module2')


@register_subcommand('module2-command', assign_args, {'add_help': False})
def command(args):
    return args.module2



# -*- coding: utf-8 -*-
"""\
- Remove `re_safe_word` column from the database, if it exists.
- Remove keywords that only contain XML.
- Strip XML from keywords, if they contain XML

"""


def up(cursor):
    # Remove `re_safe_word` column from the database, if it exists.
    cursor.execute("ALTER TABLE keywords DROP COLUMN IF EXISTS re_safe_word")
    # Remove keywords that only contain XML.
    cursor.execute("""\
    WITH kwids AS (
      DELETE FROM keywords
      WHERE word ~ '<' AND regexp_replace(word, '<.*?>', '', 'g') != ''
      RETURNING keywordid
    )
    DELETE FROM modulekeywords
    WHERE keywordid = any(SELECT keywordid FROM kwids)
    """)
    # Strip XML from keywords, if they contain XML
    cursor.execute("""\
    UPDATE keywords SET word = regexp_replace(word, '<.*?>', '', 'g')
    WHERE word ~ '<'
    """)


def down(cursor):
    pass



# -*- coding: utf-8 -*-
import argparse
try:
    from unittest import mock
except ImportError:
    import mock

import pytest
import venusian


@pytest.fixture
def parser():
    return argparse.ArgumentParser()


@pytest.fixture
def sub_parsers(parser):
    sub_parser = parser.add_subparsers()
    return mock.Mock(spec=sub_parser)


@pytest.fixture
def scanner(sub_parsers):
    return venusian.Scanner(sub_parsers=sub_parsers)


def test_registration(scanner, sub_parsers):
    from . import scanned_import
    from cnxdb.cli.discovery import SUBCOMMAND_CATEGORY as category
    scanner.scan(scanned_import, categories=(category,))

    assert sub_parsers.add_parser.call_count == 2

    sub_parsers.add_parser.assert_any_call('module1-command')
    func = scanned_import.module1.command
    sub_parsers.add_parser.assert_has_calls([
        mock.call().set_defaults(cmd=func),
    ])

    sub_parsers.add_parser.assert_any_call('module2-command', add_help=False)
    func = scanned_import.module2.command
    sub_parsers.add_parser.assert_has_calls([
        mock.call().set_defaults(cmd=func),
    ])
    sub_parsers.add_parser.assert_has_calls([
        mock.call().add_argument('module2'),
    ])


def test_discovery(parser):
    from . import scanned_import
    from cnxdb.cli.discovery import discover_subcommands
    discover_subcommands(parser, scope=scanned_import)

    args = parser.parse_args(['module1-command'])
    assert args.cmd(args)

    test_arg = 'echo'
    args = parser.parse_args(['module2-command', test_arg])
    assert args.cmd(args) == test_arg



# -*- coding: utf-8 -*-
import os
import logging
import sys
import warnings

import psycopg2

from . import exceptions
from .manifest import get_schema


here = os.path.abspath(os.path.dirname(__file__))
SCHEMA_DIR = os.path.join(here, '..', 'schema')

logger = logging.getLogger('cnxdb')


SCHEMA_INITIALIZED_CHECK = """\
DO LANGUAGE plpgsql
$$
BEGIN
  IF EXISTS (SELECT 1 FROM pg_class WHERE relname='modules') THEN
    RAISE EXCEPTION USING MESSAGE = 'Database schema is initialized.';
  END IF;
END;
$$;
"""


def _has_schema(cursor):
    """Checks for the existence of the database schema.
    Returns a boolean to indicate if the database schema exists or not.

    """
    try:
        cursor.execute(SCHEMA_INITIALIZED_CHECK)
    except psycopg2.InternalError as exc:
        if 'Database schema is initialized.' in exc.args[0]:
            return True
        else:  # pragma: no cover
            return False
    else:
        return False


def init_db(connection_string, as_venv_importable=False):
    """Initialize the database from the given ``connection_string``."""
    with psycopg2.connect(connection_string) as db_connection:
        with db_connection.cursor() as cursor:
            if _has_schema(cursor):
                raise exceptions.DBSchemaInitialized()
            for schema_part in get_schema(SCHEMA_DIR):
                cursor.execute(schema_part)
    if as_venv_importable:
        init_venv(connection_string)


ACTIVATE_VENV_SQL_FUNCTION = """\
CREATE FUNCTION venv.activate_venv()
RETURNS void LANGUAGE plpythonu AS $_$
import sys
import os
import site
old_os_path = os.environ.get('PATH','')
os.environ['PATH'] = os.path.dirname(os.path.abspath('{activate_path}')) \
+ os.pathsep + old_os_path
base = os.path.dirname(os.path.dirname(os.path.abspath('{activate_path}')))
site_packages = os.path.join(base, 'lib', 'python%s' % sys.version[:3], \
'site-packages')
prev_sys_path = list(sys.path)
site.addsitedir(site_packages)
sys.real_prefix = sys.prefix
sys.prefix = base
# Move the added items to the front of the path:
new_sys_path = []
for item in list(sys.path):
    if item not in prev_sys_path:
        new_sys_path.append(item)
        sys.path.remove(item)
sys.path[:0] = new_sys_path
$_$"""


def _is_localhost_connection(db_connection):
    """Given a database connection, check this is a connection to localhost.

    """
    # If you are connecting to a database that is not localhost,
    # don't initalize with virtualenv
    db_dict = dict(p.split('=') for p in db_connection.dsn.split())
    return db_dict.get('host', 'localhost') != 'localhost'


def init_venv(connection_string):
    """Initialize a Python virtual environment for trigger importation."""
    # If virtualenv is active, use that for postgres
    if hasattr(sys, 'real_prefix'):  # attr is only present within a venv
        activate_path = os.path.join(os.path.realpath(sys.prefix),
                                     'bin/activate_this.py')
    else:  # pragma: no cover
        return

    with psycopg2.connect(connection_string) as db_connection:
        if _is_localhost_connection(db_connection):  # pragma: no cover
            warnings.warn("An attempt to use ``init_venv`` was made, "
                          "but not on the same host as the postgres service.")
        with db_connection.cursor() as cursor:
            cursor.execute("SELECT current_database();")
            db_name = cursor.fetchone()[0]

            cursor.execute("SELECT schema_name "
                           "FROM information_schema.schemata "
                           "WHERE schema_name = 'venv';")
            try:
                schema_exists = cursor.fetchone()[0]
            except TypeError:
                cursor.execute("CREATE SCHEMA venv")
                try:
                    cursor.execute("SAVEPOINT session_preload")
                    cursor.execute("ALTER DATABASE \"{}\" SET "
                                   "session_preload_libraries ="
                                   "'session_exec'".format(db_name))
                except psycopg2.ProgrammingError as e:  # pragma: no cover
                    if e.message.startswith(
                            'unrecognized configuration parameter'):

                        cursor.execute("ROLLBACK TO SAVEPOINT "
                                       "session_preload")
                        logger.warning("Postgresql < 9.4: make sure "
                                       "to set "
                                       "'local_preload_libraries "
                                       "= session_exec' in "
                                       "postgresql.conf and restart")
                    else:  # pragma: no cover
                        raise

                cursor.execute("ALTER DATABASE \"{}\" "
                               "SET session_exec.login_name = "
                               "'venv.activate_venv'"
                               .format(db_name))
                sql = ACTIVATE_VENV_SQL_FUNCTION.format(
                    activate_path=activate_path)
                cursor.execute(sql)


__all__ = (
    'init_db',
    'init_venv',
)



# -*- coding: utf-8 -*-
import os


here = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(here, 'data')


def test_get_schema():
    example_schema_dir = os.path.join(DATA_DIR, 'example_schema')

    file_series = (
        'file_xyz.sql',
        'sub/subfile_123-xyz.sql',
        'sub/subsub/subsubfile_123-abc-456.sql',
        'sub/subsub/subsubfile_123-abc-123.sql',
        'sub/subfile_123-efg.sql',
        'file_abc.sql',
        )
    from cnxdb.init.manifest import get_schema
    for i, file in enumerate(get_schema(example_schema_dir)):
        abs_filepath = os.path.abspath(os.path.join(example_schema_dir,
                                                    file_series[i]))
        assert '-- FILE: {0}'.format(abs_filepath).encode('utf-8') in file



# -*- coding: utf-8 -*-
"""\
Manifest marker files (``manifest.json``) are placed in a schema's directory
to properly order the schema files into a linear format that can be loaded
into the database.

"""
import os
import json


SCHEMA_MANIFEST_FILENAME = 'manifest.json'


def _read_schema_manifest(manifest_filepath):
    with open(os.path.abspath(manifest_filepath), 'r') as fp:
        raw_manifest = json.load(fp)
    manifest = []
    relative_dir = os.path.abspath(os.path.dirname(manifest_filepath))
    for item in raw_manifest:
        if isinstance(item, dict):
            file = item['file']
        else:
            file = item
        if os.path.isdir(os.path.join(relative_dir, file)):
            next_manifest = os.path.join(
                relative_dir,
                file,
                SCHEMA_MANIFEST_FILENAME)
            manifest.append(_read_schema_manifest(next_manifest))
        else:
            manifest.append(os.path.join(relative_dir, file))
    return manifest


def _compile_manifest(manifest, content_modifier=None):
    """Compile a given ``manifest`` into a sequence of schema items.

    Apply the optional ``content_modifier`` to each file's contents.
    """
    items = []
    for item in manifest:
        if isinstance(item, list):
            items.extend(_compile_manifest(item, content_modifier))
        else:
            with open(item, 'r') as fp:
                content = fp.read()
            if content_modifier:
                content = content_modifier(item, content)
            items.append(content)
    return items


def get_schema(schema_directory):
    """Return the current schema."""
    manifest_filepath = os.path.join(schema_directory,
                                     SCHEMA_MANIFEST_FILENAME)
    schema_manifest = _read_schema_manifest(manifest_filepath)

    # Modify the file so that it contains comments that say it's origin.
    def file_wrapper(f, c):
        return "-- FILE: {0}\n{1}\n-- \n".format(f, c).encode('utf-8')

    return _compile_manifest(schema_manifest, file_wrapper)


__all__ = ('get_schema')



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
CREATE EXTENSION pg_trgm;
CREATE INDEX modules_strip_html_name_trgm_gin ON modules \
    USING gin(strip_html(name) gin_trgm_ops);
""")


def down(cursor):
    cursor.execute("""\
DROP INDEX IF EXISTS modules_strip_html_name_trgm_gin;
DROP EXTENSION IF EXISTS pg_trgm;
""")



# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages


IS_PY3 = sys.version_info > (3,)

setup_requires = (
    'pytest-runner',
    )
install_requires = (
    'psycopg2',
    'venusian',
    )
tests_require = [
    'pytest',
    ]
extras_require = {
    'test': tests_require,
    }
description = "Connexions Database Library"
with open('README.rst', 'r') as readme, \
     open('docs/changes.rst', 'r') as changes:
    long_description = '\n'.join([
        readme.read(),
        "==========\nChange Log\n==========",
        changes.read(),
    ])

if not IS_PY3:
    tests_require.append('mock==1.0.1')

setup(
    name='cnx-db',
    version='0.3.0',
    author='Connexions team',
    author_email='info@cnx.org',
    url="https://github.com/connexions/cnx-db",
    license='LGPL, See also LICENSE.txt',
    description=description,
    long_description=long_description,
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    test_suite='cnxdb.tests',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'cnxdb': ['*-sql/*.sql', '*-sql/**/*.sql', 'schema/*.json'],
        'cnxdb.tests': ['data/init/**/*.*'],
        },
    entry_points="""\
    [console_scripts]
    cnx-db = cnxdb.cli.main:main
    [dbmigrator]
    migrations_directory = cnxdb:migrations
    """,
    )



# -*- coding: utf-8 -*-


def up(cursor):
    # Create correct index, drop less useful one
    cursor.execute("""
    CREATE INDEX modules_uuid_txt_version_idx ON modules
    (cast(uuid as text), module_version(major_version, minor_version));
    CREATE INDEX latest_modules_uuid_txt_version_idx ON latest_modules
    (cast(uuid as text), module_version(major_version, minor_version));
    DROP INDEX modules_uuid_version_idx;
    DROP INDEX latest_modules_uuid_version_idx;
    """)


def down(cursor):
    # Drop correct index, create less useful one
    cursor.execute("""
    CREATE INDEX modules_uuid_version_idx ON modules
    (uuid, module_version(major_version, minor_version));
    CREATE INDEX latest_modules_uuid_version_idx ON latest_modules
    (uuid, module_version(major_version, minor_version));
    DROP INDEX modules_uuid_txt_version_idx;
    DROP INDEX latest_modules_uuid_txt_version_idx;
    """)



# -*- coding: utf-8 -*-
import venusian


SUBCOMMAND_CATEGORY = 'subcommands'


def register_subcommand(command_name, parser_callback=None, parser_kwargs={}):
    """Register a function as a subcommand.
    For example::

        def assign_args(parser):
            parser.add_argument('argh', action='store_true')

        @register_subcommand('pirate', assign_args)
        def command(args):
            return args.argh and 'ARGH' or 'zZz'

    """
    def wrapper(wrapped):
        def callback(scanner, name, obj):
            # Create a parser for the subcommand with the given command_name.
            parser = scanner.sub_parsers.add_parser(command_name,
                                                    **parser_kwargs)
            # Call the parser's callback, which assigns arguments.
            if parser_callback is not None:
                parser_callback(parser)
            # Assign the command's execution function to the wrapped function.
            parser.set_defaults(cmd=wrapped)
        venusian.attach(wrapped, callback, category=SUBCOMMAND_CATEGORY)
        return wrapped
    return wrapper


def discover_subcommands(parser, scope=None):
    """Discover registered subcommands. The side-effect of running this
    function is that it adds the subcommands to the given parser.
    If given, `scope` can be used to scope the discovery to a specifc module
    or package.

    """
    if scope is None:
        from .. import cli as scope
    sub_parsers = parser.add_subparsers()
    scanner = venusian.Scanner(sub_parsers=sub_parsers)
    scanner.scan(scope, categories=(SUBCOMMAND_CATEGORY,))


__all__ = (
    'discover_subcommands',
    'register_subcommand',
    'SUBCOMMAND_CATEGORY',
)



# -*- coding: utf-8 -*-
import os
import sys

import psycopg2
import pytest

from .. import testing


@pytest.mark.usefixtures('db_wipe')
def test_db_init(connection_string):
    from cnxdb.init.main import init_db
    init_db(connection_string)

    def table_name_filter(table_name):
        return (not table_name.startswith('pg_') and
                not table_name.startswith('_pg_'))

    with psycopg2.connect(connection_string) as conn:
        with conn.cursor() as cursor:
            cursor.execute("SELECT table_name "
                           "FROM information_schema.tables "
                           "ORDER BY table_name")
            tables = [table_name for (table_name,) in cursor.fetchall()
                      if table_name_filter(table_name)]

    assert 'modules' in tables
    assert 'pending_documents' in tables


@pytest.mark.usefixtures('db_wipe')
def test_db_init_called_twice(connection_string):
    from cnxdb.init.main import init_db
    init_db(connection_string)

    from cnxdb.init.exceptions import DBSchemaInitialized
    try:
        init_db(connection_string)
    except DBSchemaInitialized as exc:
        pass
    else:
        assert False, "the initialization check failed"


@pytest.mark.skipif(not testing.is_venv(), reason="not within a venv")
@pytest.mark.usefixtures('db_wipe')
def test_db_init_with_venv(connection_string):
    from cnxdb.init.main import init_db
    init_db(connection_string, True)

    with psycopg2.connect(connection_string) as conn:
        with conn.cursor() as cursor:
            cursor.execute("CREATE FUNCTION pyprefix() RETURNS text LANGUAGE "
                           "plpythonu AS $$import sys;return sys.prefix$$")
            cursor.execute("SELECT pyprefix()")
            db_pyprefix = cursor.fetchone()[0]

    assert os.path.samefile(db_pyprefix, sys.prefix)



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""\
CREATE TABLE api_keys (
  "id" SERIAL PRIMARY KEY,
  -- Any text that the service will use as an api key.
  "key" TEXT NOT NULL,
  -- This is a human readable name of the person, organization or service
  -- that we are giving an api key.
  "name" TEXT NOT NULL,
  -- A list of groups that this api key is a member of.
  -- For example, g:publishers or g:trusted-publishers.
  -- See the documenation about available groups.
  "groups" TEXT[]
);
""")


def down(cursor):
    cursor.execute("DROP TABLE api_keys;")



# -*- coding: utf-8 -*-


def up(cursor):
    cursor.execute("""
CREATE OR REPLACE FUNCTION index_collated_fulltext_trigger()
  RETURNS TRIGGER AS $$
  DECLARE
    has_existing_record integer;
    _baretext text;
    _idx_vectors tsvector;
  BEGIN
    has_existing_record := (SELECT item FROM collated_fti WHERE item = NEW.item and context = NEW.context);
    _baretext := (SELECT xml_to_baretext(convert_from(f.file, 'UTF8')::xml)::text FROM files AS f WHERE f.fileid = NEW.fileid);
    _idx_vectors := to_tsvector(_baretext);

    IF has_existing_record IS NULL THEN
      INSERT INTO collated_fti (item, context, fulltext, module_idx)
        VALUES ( NEW.item, NEW.context,_baretext, _idx_vectors );
    ELSE
      UPDATE collated_fti SET (fulltext, module_idx) = ( _baretext, _idx_vectors )
        WHERE item = NEW.item and context = NEW.context;
    END IF;
    RETURN NEW;
  END;
  $$
  LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION index_collated_fulltext_lexeme_update_trigger()
  RETURNS TRIGGER AS $$
  BEGIN

    DELETE from collated_fti_lexemes where item = NEW.item AND context = NEW.context;

    INSERT into collated_fti_lexemes (item, context, lexeme, positions)
       (with lex as (SELECT regexp_split_to_table(NEW.module_idx::text, E' \\'') as t )
       SELECT NEW.item, NEW.context,
              substring(t,1,strpos(t,E'\\':')-1),
              ('{'||substring(t,strpos(t,E'\\':')+2)||'}')::int[] from lex) ;

  RETURN NEW;
  END;
  $$
  LANGUAGE plpgsql;
            """)


def down(cursor):
    cursor.execute("""
CREATE OR REPLACE FUNCTION index_collated_fulltext_trigger()
  RETURNS TRIGGER AS $$
  DECLARE
    has_existing_record integer;
    _baretext text;
    _idx_vectors tsvector;
  BEGIN
    has_existing_record := (SELECT item, context FROM collated_fti WHERE item = NEW.item and context = NEW.context);
    _baretext := (SELECT xml_to_baretext(convert_from(f.file, 'UTF8')::xml)::text FROM files AS f WHERE f.fileid = NEW.fileid);
    _idx_vectors := to_tsvector(_baretext);

    IF has_existing_record IS NULL THEN
      INSERT INTO collated_fti (item, context, fulltext, module_idx)
        VALUES ( NEW.item, NEW.context,_baretext, _idx_vectors );
    ELSE
      UPDATE collated_fti SET (fulltext, module_idx) = ( _baretext, _idx_vectors )
        WHERE item = NEW.item and context = NEW.context;
    END IF;
    RETURN NEW;
  END;
  $$
  LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION index_collated_fulltext_lexeme_update_trigger()
  RETURNS TRIGGER AS $$
  BEGIN

    DELETE from collated_fti_lexemes where item = NEW.item;

    INSERT into collated_fti_lexemes (item, context, lexeme, positions)
       (with lex as (SELECT regexp_split_to_table(NEW.module_idx::text, E' \\'') as t )
       SELECT NEW.item, NEW.context,
              substring(t,1,strpos(t,E'\\':')-1),
              ('{'||substring(t,strpos(t,E'\\':')+2)||'}')::int[] from lex) ;

  RETURN NEW;
  END;
  $$
  LANGUAGE plpgsql;

    """)



# -*- coding: utf-8 -*-
"""\
Add more states to modulestates for post publication
"""


def up(cursor):
    cursor.execute("""\
        INSERT INTO modulestates (stateid, statename) VALUES
            (5, 'post-publication'),
            (6, 'processing'),
            (7, 'errored')""")


def down(cursor):
    cursor.execute("""\
        DELETE FROM modulestates WHERE statename IN
            ('post-publication', 'processing', 'errored')""")
    cursor.execute("""\
        SELECT setval('modulestates_stateid_seq', 5, false)""")
